/*
 *  vplay.c - plays and records 
 *            CREATIVE LABS VOICE-files, Microsoft WAVE-files and raw data
 *
 *  Autor:    	Michael Beck - beck@informatik.hu-berlin.de
 *  Recoded by:	Jaroslav Kysela - perex@jcu.cz 	
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <ctype.h>
#define __GUS_PATH_APP__
#include "../../include/libgus.h"
#include "fmtheaders.h"

#define VERSION			"1.01"

#define DEFAULT_DSP_SPEED 	8000

#define min(a,b) 		((a) <= (b) ? (a) : (b))

#define VOC_FMT			0
#define WAVE_FMT		1
#define RAW_DATA		2

/* global data */

int audio_fmt = GUS_PCM_MODE_NONE;
int timelimit = 0, dsp_speed = DEFAULT_DSP_SPEED, dsp_stereo = 0;
int samplesize   = 8;
int quiet_mode   = 0;
int verbose_mode = 0;
int record_type  = WAVE_FMT;
u_long count;
int abuf_size, zbuf_size;
int direction, omode;
char *audiobuf, *zerobuf, *command;
int vocminor, vocmajor;		/* .VOC version */

/* defaults for playing raw data */
struct {
  int timelimit, dsp_speed, dsp_stereo, samplesize;
} raw_info = { 0, DEFAULT_DSP_SPEED, 0, 8 };

/* needed prototypes */

void record_play (char *name);
void start_voc (int fd, u_long count);
void start_wave (int fd, u_long count);
void end_voc (int fd);
void end_wave_raw (int fd);

struct fmt_record {
  void (*start) (int fd, u_long count);
  void (*end) (int fd);
  char *what;
} fmt_rec_table[] = {
  { start_voc,  end_voc,      "VOC" },
  { start_wave, end_wave_raw, "WAVE" },
  { NULL,       end_wave_raw, "raw data" }
}; 

static void usage( char *command )
{
  fprintf (stderr,
"Usage: %s [switches] [filename] <filename> ...\n"
"Available switches:\n"
"\n"
"  -c <card>   select card # or card id (1-%i), defaults to 1\n"
"  -q          quiet mode\n"
"  -v          file format Voc\n"
"  -w          file format Wave\n"
"  -r          file format Raw\n"
"  -S          stereo\n"
"  -t <secs>   timelimit (seconds)\n"
"  -s <Hz>     speed (Hz)\n"
"  -b <bits>   sample size (8,16 bits)\n"
"  -B <size>   set audio buffer size\n"
"  -m          set CD-ROM quality (44100Hz,stereo,16-bit)\n"
"  -M          set DAT quality - CODEC only (48000Hz,stereo,16-bit)\n"
"  -p <type>   compression type (alaw, ulaw, adpcm)\n"
, command, gus_pcm_cards() );
}

int main (int argc, char *argv[])
{
  int card = 0;
  char c;

  command = argv[0];
  if (strstr (argv[0], "ultrarec")) {
    direction = GUS_PCM_DIRECTION_RECORD;
  }
  else if (strstr (argv[0], "ultraplay")) {
    direction = GUS_PCM_DIRECTION_PLAY;
  }
  else {
    fprintf (stderr,
             "Error: command should be named either ultrarec or ultraplay\n");
    exit (1);
  }

  abuf_size = -1;
  while ((c = getopt (argc, argv, "c:qs:St:b:vrwdB:c:p:mM")) != EOF)
    switch (c) {
      case 'c':
        card = gus_look_for_card( optarg );
        if ( card < 0 )
          {
            fprintf( stderr, "card '%s' not found\n", optarg );
            exit( -1 );
          }
        break;
      case 'S':
	dsp_stereo = raw_info.dsp_stereo = 1;
	break;
      case 'q':
	quiet_mode = 1;
	break;
      case 'r':
        record_type = RAW_DATA;
        break;
      case 'v' :
        record_type = VOC_FMT;
        break;
      case 'w' :
        record_type = WAVE_FMT;
        break;
      case 's':
	dsp_speed = atoi (optarg);
	if (dsp_speed < 300)
	  dsp_speed *= 1000;
        raw_info.dsp_speed = dsp_speed;
	break;
      case 't':
	timelimit = raw_info.timelimit = atoi (optarg);
	break;
      case 'b':
	samplesize = raw_info.samplesize = atoi (optarg);
	break;
      case 'd':
        verbose_mode = 1; quiet_mode = 0;
        break;
      case 'B':
        abuf_size = atoi (optarg) * 1024L;
        break;
      case 'p':
	if ( !strcmp( optarg, "alaw" ) )
	  {
	    audio_fmt = GUS_PCM_MODE_A_LAW;
	    samplesize = 8;
	    record_type = RAW_DATA;
	    break;
	  }
	if ( !strcmp( optarg, "ulaw" ) )
	  {
	    audio_fmt = GUS_PCM_MODE_MU_LAW;
	    samplesize = 8;
	    record_type = RAW_DATA;
	    break;
	  }
	if ( !strcmp( optarg, "adpcm" ) )
	  {
	    audio_fmt = GUS_PCM_MODE_IMA_ADPCM;
	    samplesize = 16;
	    record_type = RAW_DATA;
	    break;
	  }
        fprintf( stderr, "wrong compressed format\n" );
        exit( -1 ); 
      case 'm':
      case 'M':
	samplesize = raw_info.samplesize = 16;
	dsp_stereo = raw_info.dsp_stereo = 1;
        dsp_speed = raw_info.dsp_speed = c == 'M' ? 48000 : 44100;
	break;
      default:
        usage( command );
	exit(-1);
    }

  if ( !quiet_mode )
    printf( "%s version " VERSION " by Jaroslav Kysela & Michael Beck\n", command );
    
  if ( gus_pcm_open( card, direction, GUS_PCM_OF_NONE ) < 0 )
    {
      fprintf( stderr, "audio open error\n" );
      exit( -1 );
    }

  if ( dsp_speed < 2000 && gus_pcm_rate_supported( direction ) < dsp_speed )
    {
      fprintf( stderr, "Unsupported rate %iHz. Rate can be 2000Hz-%iHz.\n", dsp_speed, gus_pcm_rate_supported( direction ) );
      exit( -1 );
    }
  if ( samplesize == 16 && !(gus_pcm_mode_supported( direction ) & GUS_PCM_MODE_S16_LE) )
    {
      fprintf( stderr, "16-bit %s isn't supported by hardware.\n", direction == GUS_PCM_DIRECTION_PLAY ? "playback" : "recording" );
      exit( -1 );
    }
  
  if ( abuf_size < 0 )
    abuf_size = gus_pcm_dma_size_get( direction, 0 );
  if (abuf_size < 512 || abuf_size > 128L * 1024L * 1024L ) {
    if (abuf_size == -1)
      fprintf( stderr, "Can't get size of audio buffer\n" );
    else
      fprintf (stderr, "Invalid audio buffer size %d\n", abuf_size);
    exit (-1);
  }

  if ( audio_fmt != 0 )
    {
      if ( ( gus_pcm_mode_supported( direction ) & audio_fmt ) != audio_fmt ) 
        {
          fprintf(stderr, "%s: format isn't supported by audio driver\n", command);
          exit (-1);
        }
    }
  
  zbuf_size = 256;
  if ( (audiobuf = malloc (abuf_size)) == NULL ||
       (zerobuf  = malloc (zbuf_size)) == NULL )  {
    fprintf (stderr, "%s: unable to allocate input/output buffer\n", command);
    exit (-1);
  }
  memset (zerobuf, 128, zbuf_size);

  if (optind > argc - 1)
    record_play (NULL);
  else
    while (optind <= argc - 1) {
      record_play (argv[optind++]);
    }
  gus_pcm_close( card );
  return 0;
}

/*
 test, if it is a .VOC file and return >=0 if ok (this is the length of rest)
                                       < 0 if not 
*/
int test_vocfile(void *buffer)
{
  VocHeader *vp = buffer;
  if (strstr(vp->magic, MAGIC_STRING) ) {
    vocminor = vp->version & 0xFF;
    vocmajor = vp->version / 256;
    if (vp->version != (0x1233 - vp->coded_ver) )
      return -2;				/* coded version mismatch */
    return vp->headerlen - sizeof(VocHeader);	/* 0 mostly */
  }
  return -1;					/* magic string fail */
}

/*
 test, if it's a .WAV file, 0 if ok (and set the speed, stereo etc.)
                          < 0 if not
*/
int test_wavefile(void *buffer)
{
  WaveHeader *wp = buffer;
  if (wp->main_chunk == RIFF && wp->chunk_type == WAVE &&
      wp->sub_chunk == FMT && wp->data_chunk == DATA) {
    if (wp->format != PCM_CODE) {
      fprintf (stderr, "%s: can't play not PCM-coded WAVE-files\n", command);
      exit (-1);
    }
    if (wp->modus > 2) {
      fprintf (stderr, "%s: can't play WAVE-files with %d tracks\n",
               command, wp->modus);
      exit (-1);
    }
    dsp_stereo = (wp->modus == WAVE_STEREO) ? 1 : 0;
    samplesize = wp->bit_p_spl;
    dsp_speed = wp->sample_fq; 
    count = wp->data_length;
    return 0;
  }
  return -1;
}
 
/*
  writing zeros from the zerobuf to simulate silence,
  perhaps it's enough to use a long var instead of zerobuf ?
*/
void write_zeros (unsigned x)
{
  unsigned l;
 
  while (x) {
    l = min (x, zbuf_size);
    if (gus_pcm_write(zerobuf, l) != l) {
      fprintf( stderr, "write error\n" );
      exit (-1);
    }
    x -= l;
  }
} 

/* if need a SYNC, 
   (is needed if we plan to change speed, stereo ... during output)
*/
inline void sync_dsp(void)
{
  if ( gus_pcm_sync() < 0 )
    {
      fprintf( stderr, "%s: sync failed\n", command );
      exit (-1);
    }
}

/* setting the speed for output */
void set_dsp_speed (int dsp_speed)
{
  if (gus_pcm_rate_set(direction, dsp_speed) < 0) {
    fprintf (stderr, "%s: unable to set audio speed\n", command);
    exit (-1);
  }
}

/*
  ok, let's play a .voc file
*/ 
void vplay (int fd, int ofs, char *name)
{
  int l;
  BlockType *bp;
  Voice_data *vd;
  Ext_Block *eb;
  u_long nextblock, in_buffer;
  u_char *data = audiobuf;
  char was_extended = 0, output = 0;
  u_short *sp, repeat = 0;
  u_long silence;
  int filepos = 0;
  char one_chn = 0;

#define COUNT(x)	nextblock -= x; in_buffer -=x ;data += x

  /* first SYNC the dsp */
  sync_dsp();
 
  if (!quiet_mode) 
    fprintf (stderr, "Playing Creative Labs Voice file ...\n");

  /* first we waste the rest of header, ugly but we don't need seek */
  while (ofs > abuf_size) {	
    read (fd, audiobuf, abuf_size);
    ofs -= abuf_size;
  }
  if (ofs)
    read (fd, audiobuf, ofs);

  /* .voc files are 8 bit (now) */
  samplesize = VOC_SAMPLESIZE;
  if ( gus_pcm_mode_set( direction, samplesize ) < 0 )
    {
      fprintf(stderr, "%s: unable to set 8 bit sample size!\n", command);
      exit (-1);
    }
  /* and there are MONO by default */
  dsp_stereo = MODE_MONO;
  gus_pcm_channels_set( direction, dsp_stereo ? 2 : 1 );

  in_buffer = nextblock = 0;
  while (1) {
    Fill_the_buffer:		/* need this for repeat */
    if ( in_buffer < 32 ) {
      /* move the rest of buffer to pos 0 and fill the audiobuf up */
      if (in_buffer)
        memcpy (audiobuf, data, in_buffer);
      data = audiobuf;
      if ((l = read (fd, audiobuf + in_buffer, abuf_size - in_buffer) ) > 0) 
        in_buffer += l;
      else if (! in_buffer) {
	/* the file is truncated, so simulate 'Terminator' 
           and reduce the datablock for save landing */
        nextblock = audiobuf[0] = 0;
        if (l == -1) {
          perror (name);
          exit (-1);
        }
      }
    }
    while (! nextblock) { 	/* this is a new block */
      bp = (BlockType *)data; COUNT(sizeof (BlockType));
      nextblock = DATALEN(bp);
      if (output && !quiet_mode)
        fprintf (stderr, "\n");	/* write /n after ASCII-out */
      output = 0;
      switch (bp->type) {
        case 0:
#if 0
          d_printf ("Terminator\n");
#endif
          return;		/* VOC-file stop */
        case 1:
          vd = (Voice_data *)data; COUNT(sizeof(Voice_data));
          /* we need a SYNC, before we can set new SPEED, STEREO ... */
          sync_dsp();

          if (! was_extended) {
            dsp_speed = (int)(vd->tc);
            dsp_speed = 1000000 / (256 - dsp_speed); 
#if 0
            d_printf ("Voice data %d Hz\n", dsp_speed);
#endif
            if (vd->pack) {	/* /dev/dsp can't it */
              fprintf (stderr, "%s: can't play packed .voc files\n", command);
              return;
            }
            if (dsp_stereo) {	/* if we are in Stereo-Mode, switch back */
              dsp_stereo = MODE_MONO;
              gus_pcm_channels_set( direction, dsp_stereo ? 2 : 1 ); 
            }
          }
          else {		/* there was extended block */
            if (one_chn)	/* if one Stereo fails, why test another ? */
              dsp_stereo = MODE_MONO;
            else if (dsp_stereo) {	/* want Stereo */
              /* shit, my MACRO dosn't work here */
              gus_pcm_channels_set( direction, dsp_stereo ? 2 : 1 );
            }
            was_extended = 0;
          }
          set_dsp_speed (dsp_speed);
          break;
        case 2:			/* nothing to do, pure data */
#if 0
          d_printf ("Voice continuation\n");
#endif
          break;
        case 3:			/* a silence block, no data, only a count */
          sp = (u_short *)data; COUNT(sizeof(u_short));
          dsp_speed = (int)(*data); COUNT(1);
          dsp_speed = 1000000 / (256 - dsp_speed);
          sync_dsp();
          set_dsp_speed (dsp_speed);
          silence = ( ((u_long)*sp) * 1000) / dsp_speed; 
#if 0
          d_printf ("Silence for %d ms\n", (int)silence);
#endif
          write_zeros (*sp);
          break;
        case 4:			/* a marker for syncronisation, no effect */
          sp = (u_short *)data; COUNT(sizeof(u_short));
#if 0
          d_printf ("Marker %d\n", *sp); 
#endif
          break;
        case 5:			/* ASCII text, we copy to stderr */
          output = 1;
#if 0
          d_printf ("ASCII - text :\n");
#endif
          break; 
        case 6:			/* repeat marker, says repeatcount */
          /* my specs don't say it: maybe this can be recursive, but
             I don't think somebody use it */
          repeat = *(u_short *)data; COUNT(sizeof(u_short));
#if 0
          d_printf ("Repeat loop %d times\n", repeat);
#endif
          if (filepos >= 0)	/* if < 0, one seek fails, why test another */
            if ( (filepos = lseek (fd, 0, 1)) < 0 ) {
              fprintf(stderr, "%s: can't play loops; %s isn't seekable\n", 
                      command, name);
              repeat = 0;
            }
            else
              filepos -= in_buffer;	/* set filepos after repeat */
          else
            repeat = 0;
          break;
        case 7:			/* ok, lets repeat that be rewinding tape */
          if (repeat) {
            if (repeat != 0xFFFF) {
#if 0
              d_printf ("Repeat loop %d\n", repeat);
#endif
              --repeat;
            }
#if 0
            else
              d_printf ("Neverending loop\n");
#endif
            lseek (fd, filepos, 0);
            in_buffer = 0;		/* clear the buffer */
            goto Fill_the_buffer;
          }
#if 0
          else
            d_printf ("End repeat loop\n");
#endif
          break;
        case 8:			/* the extension to play Stereo, I have SB 1.0 :-( */
          was_extended = 1;
          eb = (Ext_Block *)data; COUNT(sizeof(Ext_Block));
          dsp_speed = (int)(eb->tc);
          dsp_speed = 256000000L / (65536 - dsp_speed);
          dsp_stereo = eb->mode;
          if (dsp_stereo == MODE_STEREO) 
            dsp_speed = dsp_speed >> 1;
          if (eb->pack) {     /* /dev/dsp can't it */
            fprintf (stderr, "%s: can't play packed .voc files\n", command);
            return;
          }
#if 0
          d_printf ("Extended block %s %d Hz\n", 
                    (eb->mode ? "Stereo" : "Mono"), dsp_speed);
#endif
          break;
        default:
          fprintf (stderr, "%s: unknown blocktype %d. terminate.\n", 
                   command, bp->type);
          return;
      } 		/* switch (bp->type) */
    }			/* while (! nextblock)  */
    /* put nextblock data bytes to dsp */
    l = min (in_buffer, nextblock);
    if (l) {  
      if (output && !quiet_mode)
        write (2, data, l);	/* to stderr */
      else {
        if (gus_pcm_write(data, l) != l) {
          fprintf( stderr, "write error\n" );
          exit(-1);
        }
      }
      COUNT(l);
    }
  }			/* while(1) */
}
/* that was a big one, perhaps somebody split it :-) */

/* setting the globals for playing raw data */
void init_raw_data(void)
{
  timelimit  = raw_info.timelimit;
  dsp_speed  = raw_info.dsp_speed;
  dsp_stereo = raw_info.dsp_stereo;
  samplesize = raw_info.samplesize;
}

/* calculate the data count to read from/to dsp */
u_long calc_count(void)
{
  u_long count;

  if (!timelimit)
    count = 0x7fffffff;
  else {
    count = timelimit * dsp_speed;
    if (dsp_stereo)
      count *= 2;
    if (samplesize != 8)
      count *= 2;
    if (audio_fmt == GUS_PCM_MODE_IMA_ADPCM)
      count /= 4;
  }
  return count;
}

/* write a .VOC-header */ 
void start_voc(int fd, u_long cnt)
{
  VocHeader  vh;
  BlockType  bt;
  Voice_data vd;
  Ext_Block  eb;
 
  strncpy(vh.magic,MAGIC_STRING,20);
  vh.magic[19] = 0x1A;
  vh.headerlen = sizeof(VocHeader);
  vh.version = ACTUAL_VERSION;
  vh.coded_ver = 0x1233 - ACTUAL_VERSION;

  write (fd, &vh, sizeof(VocHeader));

  if (dsp_stereo) {
    /* write a extended block */
    bt.type = 8;
    bt.datalen = 4;
    bt.datalen_m = bt.datalen_h = 0;
    write (fd, &bt, sizeof(BlockType));
    eb.tc = (u_short)(65536 - 256000000L / (dsp_speed << 1));
    eb.pack = 0;
    eb.mode = 1;
    write (fd, &eb, sizeof(Ext_Block));
  }
  bt.type = 1;
  cnt += sizeof(Voice_data);	/* Voice_data block follows */
  bt.datalen   = (u_char)  (cnt & 0xFF);
  bt.datalen_m = (u_char)( (cnt & 0xFF00) >> 8 );
  bt.datalen_h = (u_char)( (cnt & 0xFF0000) >> 16 );
  write (fd, &bt, sizeof(BlockType));
  vd.tc = (u_char)(256 - (1000000 / dsp_speed) );
  vd.pack = 0;
  write (fd, &vd, sizeof(Voice_data) );
} 

/* write a WAVE-header */
void start_wave(int fd, u_long cnt)
{
  WaveHeader wh;

  wh.main_chunk = RIFF;
  wh.length     = cnt + sizeof(WaveHeader) - 8; 
  wh.chunk_type = WAVE;
  wh.sub_chunk  = FMT;
  wh.sc_len     = 16;
  wh.format     = PCM_CODE;
  wh.modus      = dsp_stereo ? 2 : 1;
  wh.sample_fq  = dsp_speed;
#if 0
  wh.byte_p_spl = (samplesize == 8) ? 1 : 2;
  wh.byte_p_sec = dsp_speed * wh.modus * wh.byte_p_spl;
#else
  wh.byte_p_spl = wh.modus * ((samplesize + 7) / 8);
  wh.byte_p_sec = wh.byte_p_spl * dsp_speed;
#endif
  wh.bit_p_spl  = samplesize;
  wh.data_chunk = DATA;
  wh.data_length= cnt;
  write (fd, &wh, sizeof(WaveHeader));
}

/* closing .VOC */
void end_voc(int fd)
{
  char dummy = 0;		/* Write a Terminator */
  write (fd, &dummy, 1);
  if (fd != 1)
    close (fd);
}

void end_wave_raw(int fd)
{				/* only close output */
  if (fd != 1)
    close (fd);
}

/* playing/recording raw data, this proc handels WAVE files and
   recording .VOCs (as one block) */ 
void recplay (int fd, int loaded, u_long count, int rtype, char *name)
{
  int l;
  u_long c;
  char one_chn = 0;
  int tmps;

  sync_dsp();
  if (!quiet_mode) {
    fprintf (stderr, "%s %s : ", 
             (direction == GUS_PCM_DIRECTION_PLAY) ? "Playing" : "Recording",
             fmt_rec_table[rtype].what);
    if (samplesize != 8)
      fprintf(stderr, "%d bit, ", samplesize);
    fprintf (stderr, "Speed %d Hz ", dsp_speed);
    fprintf (stderr, "%s", dsp_stereo ? "Stereo" : "Mono");
    if ( audio_fmt == GUS_PCM_MODE_MU_LAW )
      fprintf( stderr, ", U-law" );
    if ( audio_fmt == GUS_PCM_MODE_A_LAW )
      fprintf( stderr, ", A-law" );
    if ( audio_fmt == GUS_PCM_MODE_IMA_ADPCM )
      fprintf( stderr, ", IMA ADPCM" );
    fprintf (stderr, "\n" );
  }

  if ( audio_fmt != GUS_PCM_MODE_NONE )
    tmps = audio_fmt;
   else
    tmps = samplesize == 16 ? GUS_PCM_MODE_S16_LE : GUS_PCM_MODE_U8;

  if ( gus_pcm_mode_set( direction, tmps ) < 0 )
    {
      fprintf(stderr, "%s: specified format isn't supported\n", command);
      exit( -1 );
    }

  if ( gus_pcm_channels_set( direction, dsp_stereo ? 2 : 1 ) < 0 ) {
    if (direction == GUS_PCM_DIRECTION_PLAY) {
      fprintf (stderr, "%s: can't play in Stereo; playing only one channel\n", 
               command);
      dsp_stereo = MODE_MONO;
      one_chn = 1;
    }
    else {
      fprintf (stderr, "%s: can't record in Stereo\n", command);
      exit (-1);
    }
  }
  set_dsp_speed (dsp_speed);
 
  if (direction == GUS_PCM_DIRECTION_PLAY) {
    while (count) {
      c = count;

      if (c > abuf_size)
        c = abuf_size;

      if ((l = read (fd, audiobuf + loaded, c - loaded)) > 0) {
        l += loaded; loaded = 0;	/* correct the count; ugly but ... */
#if 0
        sleep( 1 );
#endif
        if (gus_pcm_write(audiobuf, l) != l) {
	  fprintf( stderr, "write error\n" );
	  exit (-1);
	}
        count -= l;
      }
      else {
	if (l == -1) { 
	  perror (name);
          exit (-1);
        }
        count = 0;	/* Stop */
      }
    }			/* while (count) */
  }
  else {		/* we are recording */

    while (count) {
      c = count;
      if (c > abuf_size)
        c = abuf_size;

      if ((l = gus_pcm_read(audiobuf, c)) > 0) {
        if (write (fd, audiobuf, l) != l) {
	  perror (name);
	  exit (-1);
	}
        count -= l;
      }

      if (l == -1) {
        fprintf( stderr, "read error\n" );
        exit (-1);
      }
    }			/* While count */
  }
}

/*
  let's play or record it (record_type says VOC/WAVE/raw)
*/
void record_play(char *name)
{
  int fd, ofs;

  if (direction == GUS_PCM_DIRECTION_PLAY) {
 
    if (!name) {
      fd = 0;
      name = "stdin";
    }
    else if ((fd = open (name, O_RDONLY, 0)) == -1) {
      perror (name);
      exit (-1);
    }
    /* read the file header */
    read (fd, audiobuf, sizeof(VocHeader));
    if ( (ofs = test_vocfile (audiobuf) ) >= 0) 
      vplay (fd, ofs, name);
    else {
      /* read bytes for WAVE-header */
      read (fd,  audiobuf + sizeof(VocHeader), 
            sizeof(WaveHeader) - sizeof(VocHeader) );
      if (test_wavefile (audiobuf) >= 0)
        recplay (fd, 0, count, WAVE_FMT, name);
      else {
        /* should be raw data */
        init_raw_data();
        count = calc_count();
        recplay (fd, sizeof(WaveHeader), count, RAW_DATA, name);
      }
    }
    if (fd != 0)
      close(fd);
  }
  else {		/* recording ... */
    if (!name) {
      fd = 1;
      name = "stdout";
    }
    else {
      if ((fd = open (name, O_WRONLY | O_CREAT, 0666)) == -1) {
        perror (name);
        exit (-1);
      }
    }
    count = calc_count() & 0xFFFFFFFE;
    /* WAVE-file should be even (I'm not sure), but wasting one byte
       isn't a problem (this can only be in 8 bit mono) */
    if (fmt_rec_table[record_type].start)
      fmt_rec_table[record_type].start(fd, count);
    recplay (fd, 0, count, record_type, name);
    fmt_rec_table[record_type].end(fd);
  }
} 

